<?php
/*
 *********************************************************************************
 * class ChessBoard
 *********************************************************************************
 */

class ChessBoard	{

	/**
	 * Constructeur
	 */
	function ChessBoard() {
		$this->clearBoard();
	}

	/*
	 * Create empty board
	 */
	function clearBoard() {
		global $pgnConst;
		$this->castle_wk= 0;
		$this->castle_wq= 0;
		$this->castle_bk= 0;
		$this->castle_bq= 0;
		$this->moveNumber= 0;
		$this->enPassant= -1;
		$this->board= array();
		for ($r= 1; $r <= 8; $r++) {
			for ($f= 1; $f <= 8; $f++) {
				$this->setPiece($f . $r,new ChessPiece($pgnConst->Empty,$pgnConst->NoColor));
			}
		}
	}

	/*
	 * Copy board
	 */
	function copy(&$refBoard) {
		global $pgnConst;

		$this->castle_wk= $refBoard->castle_wk;
		$this->castle_wq= $refBoard->castle_wq;
		$this->castle_bk= $refBoard->castle_bk;
		$this->castle_bq= $refBoard->castle_bq;
		$this->moveNumber= $refBoard->moveNumber;
		$this->enPassant= $refBoard->enPassant;

		for ($r= 1; $r <= 8; $r++) {
			for ($f= 1; $f <= 8; $f++) {
				$original= &$refBoard->getPiece($f . $r);
				if ($original->type!=$pgnConst->Empty) {
					$piece= &new ChessPiece();
					$piece->copy($original);
					$this->setPiece($f . $r,$piece);
				}
			}
		}
	}

	/*
	 * get coord id
	 */
	function convCoord($coord) {
		$file= ord(substr($coord,0,1)) - ord('a') + 1;
		$rank= substr($coord,1,1);
		return ($file . $rank);
	}

	/*
	 * set position
	 */
	function setPiece($pos,&$piece) {
		$this->board[$pos] = &$piece;
		$piece->setPos($pos);
	}

	/*
	 * get position
	 */
	function &getPiece($pos) {
		$piece= &$this->board[$pos];
		if (!isset($piece)) {
			traceError("getPiece ". $pos ." = ???");
			$this->dump();
		}
		return $piece;
	}

	/*
	 * FEN to BOARD translation
	 */
	function fen2board($str) {
		global $pgnConst;
		$tokens= explode(' ',$str);
		$nb= count($tokens);

		// create an empty board
		$this->clearBoard();

		// load pieces
		$ranks= explode('/',$tokens[0]);
		for ($r= 8; $r > 0; $r--) {
			$f= 1;
			$values= $ranks[8-$r];
			$nb= strlen($values);
			for ($i= 0; $i<$nb; $i++) {
				$c= substr($values,$i,1);
				if (strpos($pgnConst->WPieces,$c) !==false ) {
					$this->setPiece($f.$r,new ChessPiece($c,$pgnConst->White));
					$f++;
				} else if (strpos($pgnConst->BPieces,$c) !==false ) {
					$this->setPiece($f.$r,new ChessPiece(strtoupper($c),$pgnConst->Black));
					$f++;
				} else {
					$f+= $c + 0;
				}
				
			}
		}

		// which color to move
		$this->toMove= strtolower($tokens[1]);

		// castles
		if (count($tokens) > 2) {
			if (strpos($tokens[2],"K") !==false ) {
				$this->castle_wk= 1;
			}
			if (strpos($tokens[2],"Q") !==false ) {
				$this->castle_wq= 1;
			}
			if (strpos($tokens[2],"k") !==false ) {
				$this->castle_bk= 1;
			}
			if (strpos($tokens[2],"q") !==false ) {
				$this->castle_bq= 1;
			}

		}

		// en passant square
		$this->enPassant= -1;
		if (count($tokens) > 3) {
			if ($tokens[3]=="-") {
				$this->enPassant= -1;
			} else {
				$this->enPassant= $tokens[3] + 0;
			}
		}

		// move number
		if (count($tokens) > 4) {
			$this->moveNumber= $tokens[4] + 0;
		}
	}

	/*
	 * roi en checs ?
	 */
	function isKingInCheck($color) {
		global $pgnConst;
		$KPos = &$this->findPiece($pgnConst->King, $color);
		$dest  = $KPos->getPos();

		/* inverser la couleur */
		if ($color==$pgnConst->White) {
			$color= $pgnConst->Black;
		} else {
			$color= $pgnConst->White;
		}

		$destF= substr($dest,0,1);
		$destR= substr($dest,1,1);
		for ($r= 1; $r <= 8; $r++) {
			for ($f= 1; $f <= 8; $f++) {
				$piece= &$this->getPiece($f.$r);
				if ($piece->color==$color) {
					/* dplacement OK ? */
					$deltaF= abs($destF - $f);
					$deltaR= abs($destR - $r);
					$ok= false;
					switch($piece->type) {
						case $pgnConst->Empty:
							break;

						case $pgnConst->King:
							break;

						case $pgnConst->Pawn:
							/*
							 * @@@  traiter
							 * mais a semble inutile
							 */
							break;

						case $pgnConst->Knight:
							if (($deltaF==1 && $deltaR==2) || ($deltaF==2 && $deltaR==1)) {
								return true;
							}
							break;

						case $pgnConst->Queen:
							if ($deltaF==0 || $deltaR==0) {
								if ($this->isEmpty($f,$r,$destF,$destR)) {
									return true;
								}
							}
							if ($deltaF==$deltaR) {
								if ($this->isEmpty($f,$r,$destF,$destR)) {
									return true;
								}
							}
							break;

						case $pgnConst->Bishop:
							if ($deltaF==$deltaR) {
								if ($this->isEmpty($f,$r,$destF,$destR)) {
									return true;
								}
							}
							break;

						case $pgnConst->Rook:
							if ($deltaF==0 || $deltaR==0) {
								if ($this->isEmpty($f,$r,$destF,$destR)) {
									return true;
								}
							}
							break;

						default:
							traceError("isKingInCheck : unknown piece '" . $piece->type ."'  color = ".$piece->color);

					}
				}
			}
		}
		return false;
	}

	/*
	 * move piece
	 */
	function movePiece($orgPos, $destPos) {
		$this->completeMovePiece($orgPos, $destPos);
	}

	function completeMovePiece($orgPos, $destPos) {
		global $pgnConst;
		$file   = substr($destPos,0,1);
		$rank   = substr($destPos,1,1);
		$piece= &$this->getPiece($orgPos);

		if (!isset($piece)) {
			traceError("completeMovePiece : $orgPos -> $destPos");
		}

		if ($piece==null)	return;

		if($piece->type == $pgnConst->Pawn) { // Test for en passant captures
			if (substr($orgPos,0,1)!=substr($destPos,0,1)) {
				// pawn capture something...
				$destPiece= &$this->getPiece($destPos);
				if ($destPiece->type==$pgnConst->Empty) {
					// nothing ? en passant capture !
					if($rank == 6) {
						$r= $rank-1;
					} else {
						$r= $rank+1;
					}
					$pawnCoord= $file . $r;
					// remove captured pawn
					$this->setPiece($pawnCoord, new ChessPiece($pgnConst->Empty,$pgnConst->NoColor));
				}
			}
		}

		// move piece to destination
		$this->setPiece($destPos,$piece);
		// remove piece from origin
		$this->setPiece($orgPos, new ChessPiece($pgnConst->Empty,$pgnConst->NoColor));
	}


	/*
	 * castle
	 */
	function moveKCastle(&$pgnMove,$color) {
		global $pgnConst;
		if ($color==$pgnConst->White) {
			$this->movePiece(51, 71);
			$this->movePiece(81, 61);
		} else {
			$this->movePiece(58, 78);
			$this->movePiece(88, 68);
		}
	}

	function moveQCastle(&$pgnMove,$color) {
		global $pgnConst;
		if ($color==$pgnConst->White) {
			$this->movePiece(51, 31);
			$this->movePiece(11, 41);
		} else {
			$this->movePiece(58, 38);
			$this->movePiece(18, 48);
		}
	}

	/*
	 * move King
	 */
	function moveKing(&$pgnMove,$mvTxt, $color) {
		global $pgnConst;
		$len = strlen($mvTxt);
		$fr = $this->convCoord(substr($mvTxt,$len-2,$len));
		$pieces = &$this->findPieces($pgnConst->King, $color,'00',$fr);
		$piece = &$pieces[0];
		if (count($pieces)!=1) {
			$pgnMove->addComment("moveKing : no King found (color=".$color.")");
			traceError("no King found (color=".$color.")");
			return;
		}
		$this->movePiece($piece->pos, $fr);
	}

	/*
	 * move Pawn
	 */
	function movePawn(&$pgnMove,$mvTxt, $color) {
		global $pgnConst;

		// promotion ?
		$pos= strpos($mvTxt,'=');
		$str= $mvTxt;
		if ($pos>0) {
			$str= substr($mvTxt,0,$pos);
		}
		$len= strlen($str);
		$destPos= $this->convCoord(substr($str,$len-2,$len));
		$file= substr($destPos,0,1);
		$rank= substr($destPos,1,1);
		if (strlen($str)==2) {
			// x9 : simple move, same file
			// find a pawn before or after (depending on color)
			if ($color==$pgnConst->White) {
				// previous square
				$rank--;
				$piece= &$this->getPiece($file . $rank);
				if ($piece==null) {
					traceError("no piece on board, searching a pawn  to ".$destPos." in ".$str." from ".$pgnMove->code);
				}
				if ($piece->type!=$pgnConst->Pawn) {
					// previous square again (first move)
					$rank--;
				}
			} else {
				// previous square
				$rank++;
				$piece= &$this->getPiece($file . $rank);
				if ($piece==null) {
					traceError("no piece on board, searching a pawn  to ".$destPos." in ".str." from ".$pgnMove->code);
				}
				if ($piece->type!=$pgnConst->Pawn) {
					// previous square again (first move)
					$rank++;
				}
			}
		} else {
			// yx9 : pawn capture a piece
			$file= ord(substr($str,0,1)) - ord('a') + 1;
			if ($color==$pgnConst->White) {
				// previous square
				$rank--;
			} else {
				// previous square
				$rank++;
			}
		}
		$orgPos= $file . $rank;
		if (substr($destPos,1,1)==1 || substr($destPos,1,1)==8) {
			// promotion
			if ($pos<0) {
				// queen ?
				$p= 'Q';
			} else {
				$p= substr($mvTxt,$pos+1,1);
			}
			$this->movePiece($orgPos, $destPos);
			$this->setPiece($destPos,new ChessPiece($p,$color));
			$pgnMove->promotePiece= $p;
		} else {
			$this->movePiece($orgPos, $destPos);
		}
	}

	function moveDirectionalPiece(&$pgnMove,$mvTxt, $color, $type) {

		$len= strlen($mvTxt);
		$destFR = $this->convCoord(substr($mvTxt,$len-2,$len));
		if(strlen($mvTxt) == 4) {
			$orgFR = $this->convCoord(substr($mvTxt,0,2));
			$this->movePiece($orgFR, $destFR);
			return;
		}
		if(strlen($mvTxt) > 2) {
			// $orgFR = $this->convCoord(substr($mvTxt,0,$len-2));
			$orgFR = substr($mvTxt,0,$len-2);
			if ($orgFR>='a' && $orgFR<='z') {
				$orgF= ord($orgFR) - ord('a') + 1;
				$orgR= 0;
			} else {
				$orgF= 0;
				$orgR= $orgFR;
			}
		} else {
			$orgFR= "";
			$orgF= 0;
			$orgR= 0;
		}

		$pieces = &$this->findPieces($type, $color,$orgF.$orgR,$destFR);
		if(count($pieces) == 1) {
			$piece= &$pieces[0];
			$this->movePiece($piece->getPos(), $destFR);
			return;
		}

		/*
		 * plusieurs pices :
		 * une seule valide, les autres mettent en checs
		 */
		$nb= 0;
		while ($nb<count($pieces)) {
			$piece= &$pieces[$nb];
			$orgFR= $piece->getPos();
			$pcedst= &$this->getPiece($destFR);
			/* effectuer le dplacement */
			$this->movePiece($orgFR, $destFR);
			if (!$this->isKingInCheck($color)) {
				/* tout va bien */
				return;
			}
			/* annuler ce dplacement */
			$this->setPiece($orgFR,$piece);
			$this->setPiece($destFR,$pcedst);
			$nb++;
		}
		traceError("no move ok for piece ".$type." = ".$pgnMove->board->moveNumber." ".$pgnMove->code);
		$this->dump();

	}



	/*
	 * Trouver une pice du bon type
	 */
	function &findPiece($type,$color) {
		global $pgnConst;
		for ($r= 1; $r <= 8; $r++) {
			for ($f= 1; $f <= 8; $f++) {
				$piece= &$this->getPiece($f.$r);
				if ($piece->type==$type && $piece->color==$color) {
					return $piece;
				}
			}
		}
		return null;
	}

	/*
	 * Trouver une pice du bon type
	 * pouvant effectuer le dplacement
	 * et provenant de la bonne case (partiellement connue)
	 */
	function &findPieces($type,$color,$org,$dest) {
		global $pgnConst;
		$destF= substr($dest,0,1);
		$destR= substr($dest,1,1);
		$orgF= substr($org,0,1);
		$orgR= substr($org,1,1);
		// trace("findPieces $type $color  $org -> $dest");
		$nb= 0;
		$result= array();
		for ($r= 1; $r <= 8; $r++) {
			for ($f= 1; $f <= 8; $f++) {
				$piece= &$this->getPiece($f.$r);
				if ($piece->type==$type && $piece->color==$color) {
					/* dplacement OK ? */
					$deltaF= abs($destF - $f);
					$deltaR= abs($destR - $r);
					$ok= false;
					if (($orgF==0 || $orgF==$f) && ($orgR==0 || $orgR==$r)) {
						switch($type) {
							case $pgnConst->Pawn:
								break;

							case $pgnConst->King:
								$ok= true;
								break;

							case $pgnConst->Knight:
								if (($deltaF==1 && $deltaR==2) || ($deltaF==2 && $deltaR==1)) $ok= true;
								break;

							case $pgnConst->Queen:
								if ($deltaF==0 || $deltaR==0) {
									if ($this->isEmpty($f,$r,$destF,$destR)) {
										$ok= true;
									}
								}
								if ($deltaF==$deltaR) {
									if ($this->isEmpty($f,$r,$destF,$destR)) {
										$ok= true;
									}
								}
								break;

							case $pgnConst->Bishop:
								if ($deltaF==$deltaR) {
									if ($this->isEmpty($f,$r,$destF,$destR)) {
										$ok= true;
									}
								}
								break;

							case $pgnConst->Rook:
								if ($deltaF==0 || $deltaR==0) {
									if ($this->isEmpty($f,$r,$destF,$destR)) {
										$ok= true;
									}
								}
								break;

							default:
								traceError("findPieces : unknown piece " . $piece->type ."  color = ".$piece->color);

						}
					}
					if ($ok) {
						$result[$nb]= &$piece;
						$nb++;
					}
				}
			}
		}
		if (count($result)==0) {
			traceError("nothing found ?");
			$this->dump();
		}

		return $result;
	}

	/*
	 * Vrifier que des cases sont vides
	 */
	function isEmpty($orgF,$orgR,$destF,$destR) {
		global $pgnConst;
		if ($orgF < $destF) {
			$sensF= 1;
		} elseif ($orgF > $destF) {
			$sensF= -1;
		} elseif ($orgF == $destF) {
			$sensF= 0;
		}
		if ($orgR < $destR) {
			$sensR= 1;
		} elseif ($orgR > $destR) {
			$sensR= -1;
		} elseif ($orgR == $destR) {
			$sensR= 0;
		}

		$orgF+= $sensF;
		$orgR+= $sensR;
		while ($orgF!=$destF || $orgR!=$destR) {
			$piece= &$this->getPiece($orgF.$orgR);
			if ($piece->type!=$pgnConst->Empty) {
				return false;
			}
			$orgF+= $sensF;
			$orgR+= $sensR;
		}

		return true;
	}

	/*
	 * make move
	 */
	function makeMove(&$pgnMove) {
		global $pgnConst;

		switch($pgnMove->type) {

			case $pgnConst->MoveKC:
				$this->moveKCastle($pgnMove,$pgnMove->color);
				break;

			case $pgnConst->MoveQC:
				$this->moveQCastle($pgnMove,$pgnMove->color);
				break;

			case $pgnConst->Pawn:
				$this->movePawn($pgnMove,$pgnMove->coords,$pgnMove->color);
				break;

			case $pgnConst->King:
				$this->moveKing($pgnMove,$pgnMove->coords,$pgnMove->color);
				break;

			case $pgnConst->Knight:
			case $pgnConst->Queen:
			case $pgnConst->Bishop:
			case $pgnConst->Rook:
				$this->moveDirectionalPiece($pgnMove,$pgnMove->coords, $pgnMove->color, $pgnMove->type);
				break;

			default:
				traceError("makeMove : unknown piece " . $pgnMove->type ."  move = ".$pgnMove->code);

		}
	}


	/*
	 * write
	 */
	function writeHtml() {
		global $pgnConst;
		$str= "";
		for ($r= 8; $r > 0; $r--) {
			for ($f= 1; $f <= 8; $f++) {
				$piece= &$this->board[($f . $r)];
				if ($piece->color==$pgnConst->White) {
					// echo writeHtml($piece->type);
					$str.= $piece->type;
				} else  if ($piece->color==$pgnConst->Black) {
					// echo writeHtml(strtolower($piece->type));
					$str.= strtolower($piece->type);
				} else {
					// echo writeHtml($pgnConst->Empty);
					$str.= $pgnConst->Empty;
				}
			}
		}
		return($str);
	}

	/*
	 * Display Board (no graphics)
	 */
	function dump() {
		global $pgnConst;
		traceString("-------------------\\");
		traceNL();
		for ($r= 8; $r > 0; $r--) {
			traceString(substr($pgnConst->globalsRanks,$r-1,1) . "|");
			for ($f= 1; $f <= 8; $f++) {
				$piece= &$this->getPiece($f.$r);
				if ($piece->color==$pgnConst->White) {
					traceString(" ".$piece->type);
				} else  if ($piece->color==$pgnConst->Black) {
					// traceString(' <span style="color: white; background-color: black" >'.strtolower($piece->type)."</span>");
					traceString(' '.strtolower($piece->type));
				} else {
					traceString(' '.strtoupper($piece->type));
				}
			}
			traceString(" |");
			traceNL();
		}
		traceString("-+-----------------|");
		traceNL();
		traceString(".|");
		for ($f= 0; $f < 8; $f++) {
			traceString(" ".substr($pgnConst->globalsFiles,$f,1));
		}
		traceString(" |");
		traceNL();
	}

}

?>
